from khayyam import JalaliDate, JalaliDatetime
from datetime import timedelta, datetime, date, time
from typing import Union, List

# todo: type hints
from avishan.misc.translation import AvishanTranslatable


class BchDatetime(object):
    def __init__(self, year: Union[int, datetime, date, JalaliDatetime, dict, 'BchDatetime'] = None, month: int = None,
                 day: int = None,
                 hour: int = None, minute: int = None,
                 second: int = None, microsecond: int = None) -> None:
        temp = None

        # todo not safe
        if not year and not month and not day and not hour and not minute and not second and not microsecond:
            temp = BchDatetime.from_bch_datetime(BchDatetime.now())
        elif isinstance(year, datetime):
            temp = BchDatetime.from_datetime(year)
        elif isinstance(year, date):
            temp = BchDatetime.from_date(year)
        elif isinstance(year, JalaliDatetime):
            temp = BchDatetime.from_jalali_datetime(year)
        elif isinstance(year, dict):
            temp = BchDatetime.from_dict(year)
        elif isinstance(year, BchDatetime):
            temp = year

        if temp:
            self.year, self.month, self.day, self.hour, self.minute, self.second, self.microsecond = \
                temp.year, temp.month, temp.day, temp.hour, temp.minute, temp.second, temp.microsecond
            return
        self.year = year
        if isinstance(month, str):
            try:
                month = int(month)
            except ValueError:
                try:
                    month = {
                        'فروردین': 1,
                        'اردیبهشت': 2,
                        'خرداد': 3,
                        'تیر': 4,
                        'مرداد': 5,
                        'شهریور': 6,
                        'مهر': 7,
                        'آبان': 8,
                        'آذر': 9,
                        'دی': 10,
                        'بهمن': 11,
                        'اسفند': 12,
                    }[month]
                except KeyError:
                    from avishan.exceptions import ErrorMessageException
                    raise ErrorMessageException(AvishanTranslatable(
                        EN=f'Unknown Month String {month}',
                        FA=f'عبارت ناشناخته ماه {month}'
                    ))
        self.month = month
        self.day = day
        self.hour = hour
        self.minute = minute
        self.second = second
        self.microsecond = microsecond

        """
        YYYY: year in 4 digit (1397) 
        YY00: last year in 2 digit (97)
        cYYY: year in complete digit (14)

        MM00: month in 2 digit (02)
        cM00: month in complete digit (2)

        DD00: day in 2 digit (09)
        cD00: day in complete digit (1)

        hh00: hour in 2 digit (04)
        ch00: hour in complete digit (3)
        mm00: minute in 2 digit (59)
        cm00: minute in complete digit (2)       

        ss00: second in 2 digit (59)
        cs00: second in complete digit (2)

        00ii: microseconds in 3 digit (999)
        0ii0: microseconds in 6 digit (999999)
        ci00: microseconds in real digit (17)
        """

    @property
    def year(self):
        return self.__year

    @year.setter
    def year(self, year):
        if year:
            self.__year = int(year)
        else:
            self.__year = 1

    @property
    def month(self):
        return self.__month

    @month.setter
    def month(self, month):
        if month:
            self.__month = int(month)
        else:
            self.__month = 1

    @property
    def day(self):
        return self.__day

    @day.setter
    def day(self, day):
        if day:
            self.__day = int(day)
        else:
            self.__day = 1

    @property
    def hour(self):
        return self.__hour

    @hour.setter
    def hour(self, hour):
        if hour:
            self.__hour = int(hour)
        else:
            self.__hour = 0

    @property
    def minute(self):
        return self.__minute

    @minute.setter
    def minute(self, minute):
        if minute:
            self.__minute = int(minute)
        else:
            self.__minute = 0

    @property
    def second(self):
        return self.__second

    @second.setter
    def second(self, second):
        if second:
            self.__second = int(second)
        else:
            self.__second = 0

    @property
    def microsecond(self):
        return self.__microsecond

    @microsecond.setter
    def microsecond(self, microsecond):
        if microsecond:
            self.__microsecond = int(microsecond)
        else:
            self.__microsecond = 0

    def to_dict(self, full=False, date_only=False):
        output = {
            'year': self.year, 'month': self.month, 'day': self.day, 'hour': self.hour, 'minute': self.minute,
            'second': self.second, 'microsecond': self.microsecond
        }
        if full:
            jalali_datetime = self.to_jalali_datetime()
            output['day_name'] = jalali_datetime.strftime("%A")
            output['month_name'] = jalali_datetime.strftime("%B")
        if date_only:
            del output['hour'], output['minute'], output['second'], output['microsecond']
        return output

    def to_unix_timestamp(self, resolution='second'):
        timestamp = self.to_datetime().timestamp()
        if resolution == 'second':
            return int(timestamp)
        elif resolution == 'millisecond':
            return int(timestamp * 1000)
        elif resolution == 'microsecond':
            return int(timestamp * 1000000)
        else:
            raise ValueError(AvishanTranslatable(
                EN='Incorrect resolution. Accepts are "second", "millisecond" and "microsecond". Default is "second".',
                FA='مقیاس نامعتبر. مقادیر قابل قبول: "second"، "millisecond" و "microsecond". مقدار پیش‌فرض "second".'
            ))

    def to_jalali_date(self):
        return JalaliDate(self.year, self.month, self.day)

    def to_jalali_datetime(self):
        return JalaliDatetime(self.year, self.month, self.day, self.hour, self.minute, self.second, self.microsecond)

    def to_date(self):
        return self.to_jalali_date().todate()

    def to_datetime(self) -> datetime:
        return self.to_jalali_datetime().todatetime()

    def to_time(self):
        return self.to_datetime().time()

    def to_str(self, format=None):
        if format is None:
            return str(self.to_jalali_datetime())
        """
        formats: 
        %Y: YYYY year
        %y: YY short year
        %m: MM month
        %d: DD day
        %A: persian weekday دوشنبه
        %a: short persian weekday د 
        %D: ۱۵ day
        %B: persian month شهریور
        %b: short persian month شه
        %N: persian year ۱۳۶۱
        %H: HH hour 24
        %I: II short hour 12 
        %M: MM minute
        %S: SS second
        %f: mmmmmm milliseconds
        %p: ق.ظ ب.ظ
        """
        return self.to_jalali_datetime().strftime(format)

    @staticmethod
    def from_bch_datetime(source):
        return BchDatetime(source.year, source.month, source.day, source.hour, source.minute, source.second,
                           source.microsecond)

    @staticmethod
    def from_dict(source: dict):
        return BchDatetime(
            source.get('year', None), source.get('month', None), source.get('day', None),
            source.get('hour', None), source.get('minute', None), source.get('second', None),
            source.get('microsecond', None),
        )

    @staticmethod
    def from_unix_timestamp(source):
        source = str(source)
        if len(source) > 10:  # todo is it true?
            other = source[10:]
            source = source[:10]
            temp = BchDatetime.from_datetime(datetime.fromtimestamp(int(source)))
            temp.microsecond = other * int(10 ** (6 - len(source)))
            return temp
        return BchDatetime.from_datetime(datetime.fromtimestamp(int(source)))

    @staticmethod
    def from_jalali_datetime(source):
        return BchDatetime(source.year, source.month, source.day, source.hour, source.minute, source.second,
                           source.microsecond)

    @staticmethod
    def from_date(source: date):
        return BchDatetime.from_jalali_datetime(
            JalaliDatetime(datetime.combine(source, time(0, 0, 0, 0)))
        )

    @staticmethod
    def from_datetime(source: datetime):
        return BchDatetime.from_jalali_datetime(JalaliDatetime(source))

    def load_time(self, time: datetime.time):
        self.hour = time.hour
        self.minute = time.minute
        self.second = time.second
        self.microsecond = time.microsecond
        return self

    @staticmethod
    def now():
        raise NotImplementedError('BchDatetime Deprecation')
        return BchDatetime.from_jalali_datetime(JalaliDatetime.now())

    def get_interval_timedelta_from(self, sooner):
        return self - sooner

    def is_today(self):
        now = BchDatetime()
        return now.year == self.year and now.month == self.month and now.day == self.day

    def cleaned_datetime(self, to_forward: bool, force_to_forward: bool = False, years: int = None, months: int = None,
                         days: int = None,
                         hours: int = None, minutes: int = None, seconds: int = None,
                         microseconds: int = None, ) -> 'BchDatetime':
        # todo: complete it
        # todo: if entered amount is 0, change it to 60, or any end
        new = BchDatetime(self)

        if force_to_forward and new.cleaned_datetime(
                to_forward, False, years, months, days, hours, minutes, seconds, microseconds
        ) == new:
            new += 1

        if to_forward:
            if new.microsecond % 1000000 != 0:
                new = new + timedelta(
                    microseconds=(1000000 - new.microsecond % 1000000))
            if new.second % 60 != 0:
                new = new + timedelta(
                    seconds=(60 - new.second % 60))
            if new.minute % minutes != 0:
                new = new + timedelta(minutes=(minutes - (new.minute % minutes)))
        else:
            if new.microsecond % 1000000 != 0:
                new = new - timedelta(
                    microseconds=new.microsecond % 1000000)
            if new.second % 60 != 0:
                new = new - timedelta(
                    seconds=new.second % 60)
            if new.minute % minutes != 0:
                new = new - timedelta(minutes=new.minute % minutes)
        return new

    def __add__(self, other: Union[int, timedelta]) -> 'BchDatetime':
        if isinstance(other, int):
            return BchDatetime.from_jalali_datetime(self.to_jalali_datetime() + timedelta(seconds=other))
        if isinstance(other, timedelta):
            return BchDatetime.from_jalali_datetime(self.to_jalali_datetime() + other)
        raise TypeError(AvishanTranslatable(
            EN='Accepted types are "int" as seconds and "timedelta"',
            FA='مقادیر قابل قبول "int" به عنوان ثانیه ها و "timedelta" هستند'
        ))

    def __sub__(self, other) -> Union[timedelta, 'BchDatetime']:
        if isinstance(other, BchDatetime):
            return self.to_jalali_datetime() - other.to_jalali_datetime()
        if isinstance(other, int):
            return BchDatetime.from_jalali_datetime(self.to_jalali_datetime() - timedelta(seconds=other))
        if isinstance(other, timedelta):
            return BchDatetime.from_jalali_datetime(self.to_jalali_datetime() - other)
        other = BchDatetime(other)
        return self.to_jalali_datetime() - other.to_jalali_datetime()

    # todo: accept more datetime types for add sub and compares
    def __lt__(self, other: 'BchDatetime'):
        return self.to_unix_timestamp('microsecond') < other.to_unix_timestamp('microsecond')

    def __le__(self, other):
        return self.to_unix_timestamp('microsecond') <= other.to_unix_timestamp('microsecond')

    def __gt__(self, other):
        return self.to_unix_timestamp('microsecond') > other.to_unix_timestamp('microsecond')

    def __ge__(self, other):
        return self.to_unix_timestamp('microsecond') >= other.to_unix_timestamp('microsecond')

    def __eq__(self, other):
        return self.to_unix_timestamp('microsecond') == other.to_unix_timestamp('microsecond')

    def __ne__(self, other):
        return self.to_unix_timestamp('microsecond') != other.to_unix_timestamp('microsecond')

    def __str__(self):
        return self.to_datetime().strftime("%Y/%m/%d %H:%M:%S")


class TimeRange(object):

    def __init__(self, start: BchDatetime, end: BchDatetime, belongs_to=None):
        self.start = start
        self.end = end
        self.belongs_to = belongs_to

    @property
    def length(self) -> int:
        return self.end.to_unix_timestamp() - self.start.to_unix_timestamp()

    def __str__(self):
        return str(self.start) + " - " + str(self.end)


class TimeRangeGroup(object):
    # todo 0.2.5: can optimize time ranges to fit together?
    def __init__(self, time_range_group: 'TimeRangeGroup' = None):
        self.time_ranges: List[TimeRange] = []
        if time_range_group:
            for time_range in time_range_group.time_ranges:
                self.time_ranges.append(TimeRange(time_range.start, time_range.end, time_range.belongs_to))
        self.order_time_range_group()

    @property
    def length(self) -> int:
        total = 0
        for time_range in self.time_ranges:
            total += time_range.length
        return total

    @property
    def longest_time_range_seconds(self) -> int:
        longest = 0
        for time_range in self.time_ranges:
            longest = max(longest, time_range.length)
        return longest

    @property
    def longest_time_range(self) -> TimeRange:
        longest = None
        for time_range in self.time_ranges:
            if not longest or time_range.length > longest.length:
                longest = time_range
        return longest

    def clean_with_time_slice(self, minutes: int):
        self.order_time_range_group()
        for time_range in self.time_ranges[:]:
            time_range.start = time_range.start.cleaned_datetime(to_forward=True, minutes=15)
            time_range.end = time_range.end.cleaned_datetime(to_forward=False, minutes=15)

            if time_range.length < minutes * 60:
                self.time_ranges.remove(time_range)
        self.order_time_range_group()

    def order_time_range_group(self):
        self.time_ranges.sort(key=lambda x: x.start)

    def __add__(self, other):
        if isinstance(other, TimeRangeGroup):
            result = TimeRangeGroup(self)
            for other_time_range in other.time_ranges:
                result += other_time_range
            return result

        elif isinstance(other, TimeRange):
            result = TimeRangeGroup(self)
            if len(result.time_ranges) > 0:
                for time_range in result.time_ranges[:]:
                    if time_range.start <= other.start < time_range.end < other.end:
                        result.time_ranges.remove(time_range)
                        result.time_ranges.append(TimeRange(time_range.start, other.end))
                    elif other.start < time_range.start < other.end <= time_range.end:
                        result.time_ranges.remove(time_range)
                        result.time_ranges.append(TimeRange(other.start, time_range.end))
                    elif other.start < time_range.start < time_range.end < other.end:
                        result.time_ranges.remove(time_range)
                        result.time_ranges.append(TimeRange(other.start, other.end))
                    else:
                        result.time_ranges.append(TimeRange(other.start, other.end))
            else:
                result.time_ranges.append(TimeRange(other.start, other.end))
            for time_range in result.time_ranges[:]:
                if time_range.length == 0:
                    result.time_ranges.remove(time_range)
            self.order_time_range_group()
            return result
        raise ValueError()

    def __sub__(self, other):
        if isinstance(other, TimeRangeGroup):
            result = TimeRangeGroup(self)
            for other_time_range in other.time_ranges:
                result -= other_time_range
            return result

        elif isinstance(other, TimeRange):
            for time_range in self.time_ranges[:]:
                if time_range.start < other.start < time_range.end <= other.end:
                    self.time_ranges.remove(time_range)
                    self.time_ranges.append(TimeRange(time_range.start, other.start))
                elif other.start <= time_range.start < other.end < time_range.end:
                    self.time_ranges.remove(time_range)
                    self.time_ranges.append(TimeRange(other.end, time_range.end))
                elif other.start <= time_range.start < time_range.end <= other.end:
                    self.time_ranges.remove(time_range)
                elif time_range.start < other.start < other.end < time_range.end:
                    self.time_ranges.remove(time_range)
                    self.time_ranges.append(TimeRange(time_range.start, other.start))
                    self.time_ranges.append(TimeRange(other.end, time_range.end))

            for time_range in self.time_ranges[:]:
                if time_range.length == 0:
                    self.time_ranges.remove(time_range)
            self.order_time_range_group()
            return self
        raise ValueError()

    def __str__(self):
        total = "["
        for time_range in self.time_ranges:
            total += str(time_range) + "\n"
        total = total[:-1] + "]"
        return total



# Generated by Django 3.0.9 on 2020-08-15 16:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('avishan', '0020_auto_20200803_1702'),
    ]

    operations = [
        migrations.AddField(
            model_name='emailkeyvalueauthentication',
            name='change_password_date',
            field=models.DateTimeField(blank=True, default=None, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='emailkeyvalueauthentication',
            name='change_password_token',
            field=models.CharField(blank=True, default=None, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='phonekeyvalueauthentication',
            name='change_password_date',
            field=models.DateTimeField(blank=True, default=None, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='phonekeyvalueauthentication',
            name='change_password_token',
            field=models.CharField(blank=True, default=None, max_length=255, null=True),
        ),
    ]



import inspect
from datetime import datetime
from io import BytesIO
from zipfile import ZipFile, ZipInfo

import stringcase
from typing import Type, Dict

import typing
from django.db import models
from djmoney.models.fields import MoneyField

from avishan.configure import get_avishan_config
from avishan.models import AvishanModel


class ChayiWriter:
    def __init__(self):
        self.packs = []
        for item in get_avishan_config().CHAYI_PROJECT_PACKAGE:
            self.packs.append([item[0], f'package {item[1]}.models;\n\n'
                                        f'import {item[1]}.constants.Constants;\n'])

        self.files = {}
        self.tab_size = 0
        for model in AvishanModel.get_models():
            if model._meta.abstract or model.export_ignore:
                continue
            self.files[model.class_name() + ".java"] = self.model_file_creator(model)

        self.model_file_predefined_models(self.files)

        self.compress_files(self.packs, self.files)

    def model_file_creator(self, model: Type[AvishanModel]) -> str:
        data = self.model_file_head()

        data += f'\n\npublic class {model.class_name()} extends Chayi' + ' {\n\n'
        self.tab_size = 4

        data += self.model_file_write_fields(model)
        # data += self.model_file_write_constructor(model)
        data += self.model_file_write_copy_constructor(model)
        data += self.model_file_write_direct_callable_methods(model)
        data += self.model_file_write_getter_setters(model)
        data += self.model_file_write_names(model)

        data += '    @Override\n' \
                '    public boolean equals(Object o) {\n' \
                '        if (this == o) return true;\n' \
                '        if (o == null || getClass() != o.getClass()) return false;\n' \
                f'        {model.class_name()} temp = ({model.class_name()}) o;\n' \
                '        return this.id == temp.id;\n' \
                '    }\n'

        data += "}"

        return data

    @staticmethod
    def compress_files(packs: typing.List['str'], files: Dict[str, str]):
        for pack in packs:
            archive = BytesIO()
            with ZipFile(archive, 'w') as zip_archive:
                for key, value in files.items():
                    file = ZipInfo(key)
                    zip_archive.writestr(file, pack[1] + value)

            with open(f'static/chayi_models_{pack[0]}.zip', 'wb+') as f:
                # noinspection PyTypeChecker
                f.write(archive.getbuffer())

            archive.close()

    @staticmethod
    def model_file_head() -> str:
        return f'import ir.coleo.chayi.Chayi;\n' + ChayiWriter.model_file_head_imports()

    def tab_before(self) -> str:
        return self.tab_size * " "

    @staticmethod
    def model_file_head_imports() -> str:
        return get_avishan_config().CHAYI_MODEL_FILE_IMPORTS

    def model_file_write_fields(self, model: Type[AvishanModel]) -> str:
        data = ''
        for field in model.get_full_fields():
            if field.name == 'id':
                continue
            data += self.model_file_write_field(model, field)

        for name, python_type in model.to_dict_added_fields:
            data = self.model_file_write_added_to_dict(data, name, python_type)

        return data + "\n"

    def model_file_write_constructor(self, model: Type[AvishanModel]) -> str:
        # todo default values
        create_method = getattr(model, 'create')
        data = "\n" + self.tab_before() + \
               f'public {model.class_name()}('
        for name, param in dict(inspect.signature(create_method).parameters.items()).items():
            if param.name == 'kwargs':
                raise ValueError(f'Create function for model {model.class_name()} not found')
            data += f'{self.model_file_write_param_type(param.annotation)} {name}, '
        if data.endswith(', '):
            data = data[:-2]
        data += ") {\n"

        for name in dict(inspect.signature(create_method).parameters.items()):
            data += self.tab_before() + f"    this.{name} = {name};\n"
        data += self.tab_before() + "}\n"
        return data

    @staticmethod
    def model_file_write_copy_constructor(model: Type[AvishanModel]) -> str:
        data = f'    public {model.class_name()} () ' + "{}\n\n"

        data += f"    public {model.class_name()} ({model.class_name()} {model.class_snake_case_name()}) " + "{\n"
        for field in model.get_full_fields():
            data += f'        this.{field.name} = {model.class_snake_case_name()}.{field.name};\n'
        return data + "    }\n\n"

    def model_file_write_direct_callable_methods(self, model: Type[AvishanModel]) -> str:
        from avishan.descriptor import DirectCallable

        data = ''
        skip = ['all', 'create', 'update', 'remove', 'get']

        data += '    public static final boolean create_token = true;\n'
        for direct_callable in model.direct_callable_methods():
            if direct_callable.name in skip:
                continue
            if not direct_callable.authenticate:
                data += f'    public static final boolean {direct_callable.name}_token = false;\n'
            else:
                data += f'    public static final boolean {direct_callable.name}_token = true;\n'

        if inspect.ismethod(getattr(model, 'create')):
            data += '    public static final boolean create_on_item = false;\n'
        else:
            data += '    public static final boolean create_on_item = true;\n'

        for direct_callable in model.direct_callable_methods():
            direct_callable: DirectCallable
            if direct_callable.name in skip:
                continue
            if inspect.ismethod(direct_callable.target):
                data += f'    public static final boolean {direct_callable.name}_on_item = false;\n'
            else:
                data += f'    public static final boolean {direct_callable.name}_on_item = true;\n'

        data += '\n\n'

        method = getattr(model, 'create')
        # request
        data += f'    public static RequestBody create_request('
        for key, value in dict(inspect.signature(method).parameters.items()).items():
            if key in ['self', 'cls', 'kwargs', 'args']:
                continue
            data += f'{self.model_file_write_param_type(value.annotation)} {key}, '
        if data.endswith(', '):
            data = data[:-2]
        data += ") {\n"
        data += '        JsonObject wrapper = new JsonObject();'
        data += '        JsonObject object = new JsonObject();\n'
        data += f'        wrapper.add("{model.class_snake_case_name()}", object);'
        data += self.model_file_write_direct_callable_method_body(method)
        data += '        object = wrapper;'
        data += '        return RequestBody.create(MediaType.parse("json"), object.toString());\n    }\n\n'
        for direct_callable in model.direct_callable_methods():
            direct_callable: DirectCallable
            if direct_callable.name in skip:
                continue
            method = direct_callable.target
            # request
            data += f'    public static RequestBody {direct_callable.name}_request('
            for key, value in dict(inspect.signature(method).parameters.items()).items():
                if key in ['self', 'cls', 'kwargs', 'args']:
                    continue
                data += f'{self.model_file_write_param_type(value.annotation)} {key}, '
            if data.endswith(', '):
                data = data[:-2]
            data += ") {\n"
            if direct_callable.name == 'create':
                data += '        JsonObject wrapper = new JsonObject();'
            data += '        JsonObject object = new JsonObject();\n'
            if direct_callable.name == 'create':
                data += f'        wrapper.add("{model.class_snake_case_name()}", object);'
            data += self.model_file_write_direct_callable_method_body(method)
            if direct_callable.name == 'create':
                data += '        object = wrapper;'
            data += '        return RequestBody.create(MediaType.parse("json"), object.toString());\n    }\n\n'

        return data

    def model_file_write_direct_callable_method_body(self, method) -> str:
        data = ''

        for key, value in dict(inspect.signature(method).parameters.items()).items():
            if key in ['kwargs']:
                continue
            data = self.write_with_type_check(value.annotation, key, data)
        return data

    def write_with_type_check(self, annotation, key, text, null_checked: bool = False):
        if key in ['self', 'cls']:
            return text
        if null_checked:
            text += '    '
        # noinspection PyUnresolvedReferences
        if annotation in [str, int, bool, float]:
            text += f'        object.add("{key}", new JsonPrimitive({key}));\n'
        elif isinstance(annotation, typing._Union) and len(annotation.__args__) == 2:
            text += f'        if ({key} != null) ' + "{\n"
            text = self.write_with_type_check(annotation.__args__[0], key, text, True)
            text += '        }'
        elif isinstance(annotation, typing.GenericMeta) and isinstance(annotation.__args__[0], typing._ForwardRef):
            text += f'        JsonArray {key}_array = new JsonArray();\n'
            text += f'        for (Integer i = 0; i < {key}.size(); i++) ' + "{\n"
            text += f'            {key}_array.add(RetrofitSingleTone.getInstance().getGson().toJsonTree({key}.get(i)));\n'
            text += '        }\n'
            text += f'        object.add("{key}", {key}_array);\n'
        elif isinstance(annotation, models.base.ModelBase) or isinstance(annotation, str):
            text += f'        if ({key} == null) object.add("{key}", null);\n'
            text += f'        else ' + '{\n'
            text += f'            JsonObject {key}_object = new JsonObject();\n'
            text += f'            {key}_object.add("id", new JsonPrimitive({key}.getId()));\n'
            text += f'            object.add("{key}", {key}_object);\n'
            text += '        }\n'
        elif annotation is datetime:
            text += f'        if ({key} == null) object.add("{key}", null);\n'
            text += f'        else ' + '{\n'
            text += f'            JsonObject {key}_object = new JsonObject();\n'
            text += f'            {key}_object.add("year", new JsonPrimitive({key}.year));\n'
            text += f'            {key}_object.add("month", new JsonPrimitive({key}.month));\n'
            text += f'            {key}_object.add("day", new JsonPrimitive({key}.day));\n'
            text += f'            {key}_object.add("hour", new JsonPrimitive({key}.hour));\n'
            text += f'            {key}_object.add("minute", new JsonPrimitive({key}.minute));\n'
            text += f'            {key}_object.add("second", new JsonPrimitive({key}.second));\n'
            text += f'            {key}_object.add("microsecond", new JsonPrimitive({key}.microsecond));\n'
            text += f'            object.add("{key}", {key}_object);\n'
            text += '        }\n'

        else:
            raise NotImplementedError(annotation)
        return text

    def model_file_write_getter_setters(self, model: Type[AvishanModel]) -> str:
        data = ''
        for field in model.get_fields():
            if field.name == 'id':
                continue
            data += \
                f'    public {self.model_file_write_field_type(model, field)} ' \
                f'{stringcase.camelcase("get_" + field.name)}' \
                + "() {" + \
                f'\n        return {field.name};' \
                '\n    }' \
                '\n' \
                f'\n    public void ' \
                f'{stringcase.camelcase("set_" + field.name)}' \
                + f"({self.model_file_write_field_type(model, field)} {field.name}) " + "{" + \
                f'\n        this.{field.name} = {field.name};' \
                '\n    }' \
                '\n' \
                '\n'
        return data

    @staticmethod
    def model_file_write_names(model: Type[AvishanModel]) -> str:
        return f'\n    public static String getPluralName() ' \
               '{' \
               f'\n        return "{model.class_plural_snake_case_name()}";' \
               '\n    }' \
               '\n' \
               '\n    public static String getSingleName() {' \
               f'\n        return "{model.class_snake_case_name()}";' \
               '\n    }\n\n'

    def model_file_write_field(self, model: [AvishanModel], field: models.Field) -> str:
        data = self.tab_before() + '@Expose'
        if model.chayi_ignore_serialize_field(field):
            data += '(serialize = false)'
        data += '\n' + self.tab_before() + "protected " + \
                self.model_file_write_field_type(model, field) + " " + field.name + ";\n"

        return data

    def model_file_write_field_type(self, model: Type[AvishanModel], field: models.Field) -> str:

        if isinstance(field, (models.AutoField, models.IntegerField, MoneyField)):
            return 'Integer'
        if isinstance(field, models.ForeignKey):
            return field.related_model.__name__
        if isinstance(field, models.DateTimeField):
            return 'DateTime'
        if isinstance(field, (models.CharField, models.TextField)):
            return 'String'
        if isinstance(field, models.BooleanField):
            return 'boolean'
        if isinstance(field, models.FloatField):
            return 'double'
        if isinstance(field, models.ImageField):
            return 'Image'
        if isinstance(field, models.FileField):
            return 'File'
        raise NotImplementedError()

    # noinspection PyUnresolvedReferences
    def model_file_write_param_type(self, annotation: inspect.Parameter) -> str:
        if isinstance(annotation, models.base.ModelBase):
            return annotation.__name__
        if annotation is bool:
            return 'boolean'
        if annotation is str:
            return 'String'
        if annotation is int:
            return 'Integer'
        if annotation is float:
            return 'double'
        if annotation is datetime:
            return 'DateTime'
        if isinstance(annotation, str):
            return annotation
        if isinstance(annotation, typing._Union) and len(annotation.__args__) == 2:
            return self.model_file_write_param_type(annotation.__args__[0])
        if isinstance(annotation, typing.GenericMeta):
            return f'ArrayList<{self.model_file_write_param_type(annotation.__args__[0])}>'
        if isinstance(annotation, typing._ForwardRef):
            return annotation.__forward_arg__
        raise NotImplementedError()

    @staticmethod
    def model_file_predefined_models(files: dict):
        files['DateTime.java'] = """import ir.coleo.chayi.Chayi;

import com.google.gson.annotations.Expose;

public class DateTime {

    
    @Expose(serialize = false)
    public String day_name;
    @Expose(serialize = false)
    public String month_name;
    @Expose
    public Integer year;
    @Expose
    public Integer month;
    @Expose
    public Integer day;
    @Expose
    public Integer hour;
    @Expose
    public Integer minute;
    @Expose
    public Integer second;
    @Expose
    public Integer microsecond;
    
    public DateTime() {
    
    }
    
    public DateTime(DateTime datetime) {
        this.day_name = datetime.day_name;   
        this.month_name = datetime.month_name; 
        this.year = datetime.year;   
        this.month = datetime.month;  
        this.day = datetime.day;
        this.hour = datetime.hour;   
        this.minute = datetime.minute; 
        this.second = datetime.second; 
        this.microsecond = datetime.microsecond;
    }

}"""
        files['Date.java'] = """import ir.coleo.chayi.Chayi;

import com.google.gson.annotations.Expose;

public class Date {

    @Expose(serialize = false)
    public String day_name;
    @Expose(serialize = false)
    public String month_name;
    @Expose
    public Integer year;
    @Expose
    public Integer month;
    @Expose
    public Integer day;
                                  
}"""
        files['Time.java'] = """import ir.coleo.chayi.Chayi;

import com.google.gson.annotations.Expose;

public class Time {

    @Expose
    public Integer hour;
    @Expose
    public Integer minute;
    @Expose
    public Integer second;
    @Expose
    public Integer microsecond;

}"""
        files['Image.java'] = """import ir.coleo.chayi.Chayi;

import com.google.gson.annotations.Expose;

public class Image {

    @Expose(serialize = false)
    String file;
    @Expose
    Integer id;
    
    public Integer getId() {
        return id;
    }
    public String getFile() {
        return Constants.BASE_URL + file;
    }
    
}"""
        files['File.java'] = """import ir.coleo.chayi.Chayi;

import com.google.gson.annotations.Expose;

public class File {

    @Expose(serialize = false)
    String file;
    @Expose
    Integer id;
    
    public Integer getId() {
        return id;
    }
    
    public String getFile() {
        return Constants.BASE_URL + file;
    }

}"""

    # noinspection PyUnresolvedReferences
    def model_file_write_added_to_dict(self, data: str, name: str, python_type: Type[type]) -> str:

        data += '    @Expose(serialize = false)\n'
        if python_type == int:
            writing_type = 'Integer'
        elif python_type == str:
            writing_type = 'String'
        elif issubclass(python_type, typing.List) and isinstance(python_type.__args__[0], typing._ForwardRef):
            writing_type = f'ArrayList<{python_type.__args__[0].__forward_arg__}>'
        else:
            raise NotImplementedError()

        data += f'    public {writing_type} {name};\n'
        return data



from requests import post, Response
from avishan.configure import get_avishan_config

FIREBASE_SERVER_TOKEN = get_avishan_config().FIREBASE_SERVER_TOKEN


def send_firebase_data_message(data: dict, to_key: str, server_key: str = FIREBASE_SERVER_TOKEN) -> Response:
    return post(
        url='https://fcm.googleapis.com/fcm/send',
        json={
            "data": data,
            'to': to_key,
            "android": {
                "priority": "high"
            },
        },
        headers={
            'Authorization': f'key={server_key}'}

    )


def send_firebase_notification(title: str, body: str, to_key: str, server_key: str = FIREBASE_SERVER_TOKEN):
    return post(
        url='https://fcm.googleapis.com/fcm/send',
        json={
            "notification": {
                "title": title,
                "body": body
            },
            'to': to_key,
            "android": {
                "priority": "high"
            },
        },
        headers={
            'Authorization': f'key={server_key}'}

    )


def send_firebase_data_and_notification(
        title: str,
        body: str,
        data: str,
        to_key: str,
        server_key: str = FIREBASE_SERVER_TOKEN,
        print_data: bool = False):
    data = {
            "notification": {
                "title": title,
                "body": body
            },
            'data': data,
            'to': to_key,
            "android": {
                "priority": "high"
            },
        }
    if print_data:
        print(data)
    return post(
        url='https://fcm.googleapis.com/fcm/send',
        json=data,
        headers={
            'Authorization': f'key={server_key}'}

    )



# Generated by Django 3.0.8 on 2020-08-01 19:07

import avishan.libraries.faker
import avishan.models_extensions
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('avishan', '0016_auto_20200726_1113'),
    ]

    operations = [
        migrations.CreateModel(
            name='AuthenticationVerification',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.CharField(max_length=255)),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('tried_codes', models.TextField(blank=True, null=True)),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model, avishan.libraries.faker.AvishanFaker, avishan.models_extensions.AvishanModelDjangoAdminExtension, avishan.models_extensions.AvishanModelModelDetailsExtension, avishan.models_extensions.AvishanModelFilterExtension),
        ),
        migrations.CreateModel(
            name='EmailKeyValueAuthentication',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_used', models.DateTimeField(blank=True, help_text='Last time user sent request', null=True)),
                ('last_login', models.DateTimeField(blank=True, help_text='Last time user logged in', null=True)),
                ('last_logout', models.DateTimeField(blank=True, help_text='Last time user logged out', null=True)),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('hashed_password', models.CharField(blank=True, max_length=255, null=True)),
                ('is_verified', models.BooleanField(default=False)),
                ('key', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='key_value_authentications', to='avishan.Email')),
                ('user_user_group', models.OneToOneField(help_text='Target UserUserGroup', on_delete=django.db.models.deletion.CASCADE, related_name='emailkeyvalueauthentication', to='avishan.UserUserGroup')),
                ('verification', models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='avishan.AuthenticationVerification')),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model, avishan.libraries.faker.AvishanFaker, avishan.models_extensions.AvishanModelDjangoAdminExtension, avishan.models_extensions.AvishanModelModelDetailsExtension, avishan.models_extensions.AvishanModelFilterExtension),
        ),
        migrations.CreateModel(
            name='EmailOtpAuthentication',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_used', models.DateTimeField(blank=True, help_text='Last time user sent request', null=True)),
                ('last_login', models.DateTimeField(blank=True, help_text='Last time user logged in', null=True)),
                ('last_logout', models.DateTimeField(blank=True, help_text='Last time user logged out', null=True)),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('key', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='otp_authentication', to='avishan.Email')),
                ('user_user_group', models.OneToOneField(help_text='Target UserUserGroup', on_delete=django.db.models.deletion.CASCADE, related_name='emailotpauthentication', to='avishan.UserUserGroup')),
                ('verification', models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='avishan.AuthenticationVerification')),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model, avishan.libraries.faker.AvishanFaker, avishan.models_extensions.AvishanModelDjangoAdminExtension, avishan.models_extensions.AvishanModelModelDetailsExtension, avishan.models_extensions.AvishanModelFilterExtension),
        ),
        migrations.CreateModel(
            name='PhoneKeyValueAuthentication',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_used', models.DateTimeField(blank=True, help_text='Last time user sent request', null=True)),
                ('last_login', models.DateTimeField(blank=True, help_text='Last time user logged in', null=True)),
                ('last_logout', models.DateTimeField(blank=True, help_text='Last time user logged out', null=True)),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('hashed_password', models.CharField(blank=True, max_length=255, null=True)),
                ('is_verified', models.BooleanField(default=False)),
                ('key', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='key_value_authentications', to='avishan.Phone')),
                ('user_user_group', models.OneToOneField(help_text='Target UserUserGroup', on_delete=django.db.models.deletion.CASCADE, related_name='phonekeyvalueauthentication', to='avishan.UserUserGroup')),
                ('verification', models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='avishan.AuthenticationVerification')),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model, avishan.libraries.faker.AvishanFaker, avishan.models_extensions.AvishanModelDjangoAdminExtension, avishan.models_extensions.AvishanModelModelDetailsExtension, avishan.models_extensions.AvishanModelFilterExtension),
        ),
        migrations.CreateModel(
            name='PhoneOtpAuthentication',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_used', models.DateTimeField(blank=True, help_text='Last time user sent request', null=True)),
                ('last_login', models.DateTimeField(blank=True, help_text='Last time user logged in', null=True)),
                ('last_logout', models.DateTimeField(blank=True, help_text='Last time user logged out', null=True)),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('key', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='otp_authentication', to='avishan.Phone')),
                ('user_user_group', models.OneToOneField(help_text='Target UserUserGroup', on_delete=django.db.models.deletion.CASCADE, related_name='phoneotpauthentication', to='avishan.UserUserGroup')),
                ('verification', models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='avishan.AuthenticationVerification')),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model, avishan.libraries.faker.AvishanFaker, avishan.models_extensions.AvishanModelDjangoAdminExtension, avishan.models_extensions.AvishanModelModelDetailsExtension, avishan.models_extensions.AvishanModelFilterExtension),
        ),
        migrations.CreateModel(
            name='VisitorKeyAuthentication',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_used', models.DateTimeField(blank=True, help_text='Last time user sent request', null=True)),
                ('last_login', models.DateTimeField(blank=True, help_text='Last time user logged in', null=True)),
                ('last_logout', models.DateTimeField(blank=True, help_text='Last time user logged out', null=True)),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('key', models.CharField(max_length=255)),
                ('user_user_group', models.OneToOneField(help_text='Target UserUserGroup', on_delete=django.db.models.deletion.CASCADE, related_name='visitorkeyauthentication', to='avishan.UserUserGroup')),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model, avishan.libraries.faker.AvishanFaker, avishan.models_extensions.AvishanModelDjangoAdminExtension, avishan.models_extensions.AvishanModelModelDetailsExtension, avishan.models_extensions.AvishanModelFilterExtension),
        ),
        migrations.RemoveField(
            model_name='emailverification',
            name='identifier',
        ),
        migrations.RemoveField(
            model_name='phoneotpauthenticate',
            name='key',
        ),
        migrations.RemoveField(
            model_name='phoneotpauthenticate',
            name='user_user_group',
        ),
        migrations.RemoveField(
            model_name='phonepasswordauthenticate',
            name='key',
        ),
        migrations.RemoveField(
            model_name='phonepasswordauthenticate',
            name='user_user_group',
        ),
        migrations.RemoveField(
            model_name='phoneverification',
            name='identifier',
        ),
        migrations.RemoveField(
            model_name='visitorkey',
            name='user_user_group',
        ),
        migrations.DeleteModel(
            name='EmailPasswordAuthenticate',
        ),
        migrations.DeleteModel(
            name='EmailVerification',
        ),
        migrations.DeleteModel(
            name='PhoneOTPAuthenticate',
        ),
        migrations.DeleteModel(
            name='PhonePasswordAuthenticate',
        ),
        migrations.DeleteModel(
            name='PhoneVerification',
        ),
        migrations.DeleteModel(
            name='VisitorKey',
        ),
    ]
